/*
 *  Copyright 2010 Rackspace US, Inc.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "log.h"
#include "spool.h"
#include "xen.h"

/*
 * Global variables
 */

char *dispatcher = NULL;
int agent_got_sigterm = 0;

/*
 * Private function prototypes
 */
static void _agent_signal_handler(int signum);

/*
 * Private functions
 */
static void _agent_signal_handler(int signum)
{
    switch(signum) {
        case SIGCHLD:
            while (waitpid(-1, NULL, WNOHANG) > 0) {
            }
            break;

        case SIGPIPE:
        case SIGHUP:
            break;
        case SIGTERM:
            agent_got_sigterm = 1;
            break;
        default:
            break;
    }
}

/*
 * Public functions
 */
void read_from_spool_and_send_to_xen(const char *msg) {
    char *buf = NULL, xs_path[255];
    const char *msgid;
    struct stat sb;
    int fd;

    msgid = strrchr(msg, '/');
    if (!msgid)
        msgid = msg;

    if (stat(msg, &sb) == -1)
        return;

    fd = open(msg, O_RDONLY);
    buf = mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED, fd, 0);
    snprintf(xs_path, 255, XS_MSG_OUTGOING_BASE_PATH "/%s", msgid);
    xen_write_path(xs_path, buf, sb.st_size);
    munmap(buf, sb.st_size);
    return;
}

int store_in_spool_and_call_dispatcher(const char *xs_path,
                                       const char *buf,
                                       const size_t buflen,
                                       const char *token) {
    const char *msgid = strrchr(xs_path, '/');

    if (!msgid)
        return -1;
    else
        msgid++;

    spool_store(msgid, buf, buflen);

    /* TODO: Set up a SIGCHLD handler */
    if (!fork())
        execl(dispatcher, dispatcher, msgid, NULL);
    return 0;
}


int agent_setup_signals(void)
{
    struct sigaction sa;

    memset(&sa, 0, sizeof(sa));
    sa.sa_handler = _agent_signal_handler;

    sigaction(SIGCHLD, &sa, NULL);
    sigaction(SIGPIPE, &sa, NULL);
    sigaction(SIGHUP, &sa, NULL);
    sigaction(SIGTERM, &sa, NULL);

    return 0;
}

int agent_init(const char *dispatcher_cmd) {
    AGENT_LOG0("Starting Agent Smith");

    if (agent_setup_signals() < 0) {
        AGENT_LOG0("Failed to initialise signal handlers");
        return -1;
    }

    if (xen_init() < 0) {
        AGENT_LOG0("Failed to initialise Xen connection");
        return -1;
    }

    if (spool_init() < 0) {
        AGENT_LOG0("Failed to initialise spool");
        return -1;
    }

    if (spool_watch(read_from_spool_and_send_to_xen) < 0) {
        AGENT_LOG0("Failed to register spool watch.");
        return -1;
    }

    if (xen_register_watch(XS_MSG_INCOMING_BASE_PATH,
                           store_in_spool_and_call_dispatcher,
                           "notoken") < 0) {
        AGENT_LOG0("Failed to register Xen watch.");
        return -1;
    }

    if ((dispatcher = strdup(dispatcher_cmd)) == NULL) {
        AGENT_LOG("Failed to strdup: %s", strerror(errno));
        return -1;
    }

    return 0;
}

int agent_loop(void)
{
    spool_process_pending_events();
    xen_process_pending_events();
    if (agent_got_sigterm)
        return 1;
    return 0;
}

