/*
 *  Copyright 2010 Rackspace US, Inc.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <openssl/dh.h>
#include <openssl/bn.h>

char *dh_compute_shared_key(const char *peer_key_str, DH *dh) {
    BIGNUM *peer_key, *shared_key;
    unsigned char *kbuf;
    int klen;
    char *tmp, *tmp2;

    peer_key = BN_new();
    shared_key = BN_new();

    if (BN_dec2bn(&peer_key, peer_key_str) == 0)
        return NULL;

    kbuf = malloc(DH_size(dh));
    klen = DH_compute_key(kbuf, peer_key, dh);

    BN_bin2bn(kbuf, klen, shared_key);


    tmp = BN_bn2dec(shared_key);
    tmp2 = strdup(tmp);

    free(kbuf);
    OPENSSL_free(tmp);
    BN_free(peer_key);
    BN_free(shared_key);

    return tmp2;
}

DH *dh_init(const char *p_str, const char *g_str, const char *priv_key_str, const char *pub_key_str) {
    DH *dh;

    if (((dh = DH_new()) == NULL) ||
         (BN_dec2bn(&dh->p, p_str) == 0) ||
         (BN_dec2bn(&dh->g, g_str) == 0))
        return NULL;

    if (!pub_key_str && !priv_key_str)  {
         if (DH_generate_key(dh) == 0)
            return NULL;
    } else
         if ((BN_dec2bn(&dh->priv_key, priv_key_str) == 0) ||
             (BN_dec2bn(&dh->pub_key, pub_key_str) == 0))
             return NULL;


    return dh;
}

int dh_get_shared_secret(unsigned char *buf, BIGNUM *pub_key, DH *dh) {
    return DH_compute_key(buf, pub_key, dh);
}

void usage(void) {
    printf("Usage:\n"
           "\n"
           "    dh init <statefile> <p> <g>\n"
           "or\n"
           "\n"
           "    dh compute <statefile> <peer_public_key>\n"
           "\n"
           "\n"
           "dh init <statefile> <p> <g> - Initialise diffie-hellman parameters using prime\n"
           "                              <p> and generator <g> and store them in\n"
           "                              <statefile>.\n"
           "dh compute <statefile> <k>  - Compute shared secret based on parameters in\n"
           "                              <statefile> and peer's public key <k>.\n");

}

#define write_dh_param_to_fp(dh, fp, k) write_bn_to_fp(#k, dh->k, fp);

int write_bn_to_fp(const char *key, BIGNUM *bn, FILE *fp) {
    char *tmp;
    int retval;

    retval = fprintf(fp, "%s: %s\n", key, (tmp = BN_bn2dec(bn)));
    OPENSSL_free(tmp);
    return retval;
}

int main(int argc, char *argv[]) {
    FILE *fp;

    if (argc < 4) {
        usage();
        return 1;
    }

    if (strcmp(argv[1], "init") == 0) {
        mode_t old_umask;
        DH *dh;

        if (argc != 5) {
            usage();
            return 1;
        }

        old_umask = umask(077);
        fp = fopen(argv[2], "w");
        if (!fp) {
            perror("Error opening statefile");
            return 1;
        }

        dh = dh_init(argv[3], argv[4], NULL, NULL);

        write_dh_param_to_fp(dh, fp, p);
        write_dh_param_to_fp(dh, fp, g);
        write_dh_param_to_fp(dh, fp, priv_key);
        write_dh_param_to_fp(dh, fp, pub_key);

        fclose(fp);
    } else if (strcmp(argv[1], "compute") == 0) {
        char *p_str = malloc(50);
        char *g_str = malloc(50);
        char *priv_key_str = malloc(50);
        char *pub_key_str = malloc(50);
        DH *dh;

        if (argc != 4) {
            usage();
            return 1;
        }
        fp = fopen(argv[2], "r");
        if (!fp) {
            perror("Error opening statefile");
            return 1;
        }
        if (fscanf(fp, "p: %s\n", p_str) != 1) {
            fprintf(stderr, "Failed to read statefile properly.\n");
            return 1;
        }
        if (fscanf(fp, "g: %s\n", g_str) != 1) {
            fprintf(stderr, "Failed to read statefile properly.\n");
            return 1;
        }
        if (fscanf(fp, "priv_key: %s\n", priv_key_str) != 1) {
            fprintf(stderr, "Failed to read statefile properly.\n");
            return 1;
        }
        if (fscanf(fp, "pub_key: %s\n", pub_key_str) != 1) {
            fprintf(stderr, "Failed to read statefile properly.\n");
            return 1;
        }
        if ((dh = dh_init(p_str, g_str, priv_key_str, pub_key_str)) == NULL) {
            fprintf(stderr, "Failed to read statefile properly.\n");
            return 1;
        }

        char *key;

        key = dh_compute_shared_key(argv[3], dh);
        printf("key: %s\n", key);
    } else {
        usage();
        return 1;
    }
    return 0;
}
