/*
 *  Copyright 2010 Rackspace US, Inc.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
#include <malloc.h>
#include <assert.h>
#include "slist.h"

struct _SListElem {
    void *sle_data;
    SListElem *sle_prev;
    SListElem *sle_next;
};

struct _SList {
    slist_destroy_cb sl_destroy_cb;
    SListElem sl_elem;
};

/**
 * Create a new list for storing arbitrary data
 *
 * @param destroy_cb Callback to call when deiniting (Can be NULL)
 * @retval A new list that is init'd as empty
 */
SList *slist_init(slist_destroy_cb destroy_cb)
{
    SList *list = malloc(sizeof(SList));

    if (list == NULL)
        return NULL;

    list->sl_destroy_cb = destroy_cb;
    list->sl_elem.sle_data = NULL;
    list->sl_elem.sle_prev = &(list->sl_elem);
    list->sl_elem.sle_next = &(list->sl_elem);

    return list;
}

/**
 * Destroy a previously created list
 *
 * @param slist The list to destroy
 */
void slist_deinit(SList *slist)
{
    if (slist != NULL)
    {
        SListElem *elem;
        SListElem *next;

        for(elem = slist_first(slist);
                elem != NULL;
                elem = next)
        {
            next = slist_next(slist, elem);
            if ((elem->sle_data != NULL) && slist->sl_destroy_cb != NULL)
                slist->sl_destroy_cb(elem->sle_data);
            free(elem);
        }

        free(slist);
    }
}

/**
 * Append a new element to a previously created slist
 *
 * @param list The list to append to
 * @param data Arbitrary data pointer
 * @retval The newly created element
 */
SListElem *slist_append(SList *list, void *data) {
    SListElem *elem;

    assert(list != NULL);

    elem = malloc(sizeof(SListElem));
    if (elem == NULL)
        return NULL;

    elem->sle_data = data;

    elem->sle_prev = list->sl_elem.sle_prev;
    elem->sle_next = &(list->sl_elem);
    list->sl_elem.sle_prev->sle_next = elem;
    list->sl_elem.sle_prev = elem;

    return elem;
}

/**
 * Return the first element of a previously created list
 *
 * @param list The list
 * @retval The first element or NULL if none
 */
SListElem *slist_first(SList *list) {
    return slist_next(list, &(list->sl_elem));
}

/**
 * Return the last element of a previously created list
 *
 * @param list The list
 * @retval The last element or NULL if none
 */
SListElem *slist_last(SList *list) {
    return slist_prev(list, &(list->sl_elem));
}

/**
 * Return the next element in the list after the passed element
 *
 * @param list The list
 * @param elem The current element
 * @retval The element after the current element or NULL if none left
 */
SListElem *slist_next(SList *list, SListElem *elem) {
    return elem->sle_next == &(list->sl_elem) ? NULL : elem->sle_next;
}

/**
 * Return the previous element in the list after the passed element
 *
 * @param list The list
 * @param elem The current element
 * @retval The element before the current element or NULL if none
 */
SListElem *slist_prev(SList *list, SListElem *elem) {
    return elem->sle_prev == &(list->sl_elem) ? NULL : elem->sle_prev;
}

/**
 * Return the arbitrary data pointer for an element
 *
 * @param elem The element you want the data for
 * @retval The arbitrary data pointer
 */
void *slist_data(SListElem *elem) {
    return elem->sle_data;
}

