/*
 *  Copyright 2010 Rackspace US, Inc.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
#include <sys/inotify.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "log.h"
#include "slist.h"
#include "spool.h"
#include "util.h"

#ifndef SPOOLDIR
#define SPOOLDIR "/var/spool/agent"
#endif

#define INCOMING_DIR "incoming"
#define OUTGOING_DIR "outgoing"

/*
 * Global variables
 */

int inotify_fd = -1;
SList *callbacks = NULL;

int spool_ensure_dir(const char *dir);

char *spool_basedir(void);
static char *spool_queue_dir(int in);

int spool_init(void) {

    char *tmp;

    callbacks = slist_init(NULL);
    if (callbacks == NULL)
    {
        return -1;
    }

    if (spool_ensure_dir(tmp = spool_incoming_dir()) == 0)
        free(tmp);
    else {
        syslog(LOG_DAEMON | LOG_CRIT, "Failed to create incoming spool dir: %s", tmp);
        free(tmp);
        slist_deinit(callbacks);
        callbacks = NULL;
        return -1;
    }

    if (spool_ensure_dir(tmp = spool_outgoing_dir()) == 0)
        free(tmp);
    else {
        syslog(LOG_DAEMON | LOG_CRIT, "Failed to create outgoing spool dir: %s", tmp);
        free(tmp);
        slist_deinit(callbacks);
        callbacks = NULL;
        return -1;
    }

    return 0;
}

/**
 * Get the base spool dir
 *
 * This is set at ./configure time, but can be overriden by the
 * AGENT_MSG_SPOOL environment variable.
 *
 * @return A statically allocated string denoting the base spool dir
 */
char *spool_basedir(void) {
    if (getenv("AGENT_MSG_SPOOL"))
        return getenv("AGENT_MSG_SPOOL");
    else
        return SPOOLDIR;
}

/**
 * Get the directory for incoming or outgoing messages
 *
 * @param in return directory for incoming message?
 * @retval a dynamically allocated string denoting the directory for outgoing
 *         messages if in == 0, otherwise for incoming messages
 */
static char *spool_queue_dir(int in) {
    char *basedir;
    char *retval;

    basedir = spool_basedir();
    retval = util_join_paths(basedir, in ? INCOMING_DIR : OUTGOING_DIR);
    if (retval == NULL)
        return NULL;
    if (spool_ensure_dir(retval) < 0) {
        free(retval);
        return NULL;
    }
    return retval;
}

/**
 * Get the directory for incoming messages
 *
 * @return A dynamically allocated string denoting the directory for
 *         incoming messages
 */
char *spool_incoming_dir(void) {
    return spool_queue_dir(1);
}

/**
 * Get the directory for outgoing messages
 *
 * @return A dynamically allocated string denoting the directory for
 *         outgoing messages
 */
char *spool_outgoing_dir(void) {
    return spool_queue_dir(0);
}

/**
 * Ensure existence of a directory
 *
 * Checks whether the given directory exists, and if not, attempt to
 * create it.
 *
 * @param dir The directory whose existence will be ensured
 * @return -1 in case of failure, 0 otherwise
 */
int spool_ensure_dir(const char *dir) {
    struct stat dstat;

    if (stat(dir, &dstat) == 0) {
        if (S_ISDIR(dstat.st_mode))
            return 0;
        else {
            syslog(LOG_DAEMON | LOG_CRIT, "'%s' exists, but is not a directory.", dir);
            return -1;
        }
    } else {
        if (errno == ENOENT) {
            char *tmp, *parent;
            if ((tmp = strdup(dir)) == NULL)
                return -1;

            parent = dirname(tmp);
            if (spool_ensure_dir(parent) < 0) {
                free(tmp);
                return -1;
            }
            free(tmp);
                
            if (mkdir(dir, 0700) < 0) {
                syslog(LOG_DAEMON | LOG_CRIT, "mkdir '%s' failed: %s", dir, strerror(errno));
                return -1;
            } else {
                return 0;
            }
        } else {
            syslog(LOG_DAEMON | LOG_CRIT, "stat('%s') failed: %s", dir, strerror(errno));
            return -1;
        }
    }
}

/**
 * Write a buffer to a file
 *
 * @param buf Pointer to the data to be written to the file
 * @param buflen Length of the data to be written to the file
 * @param filepath Path to the target file which must not exist
 * @return -1 if the file already exists or if an error occurs while writing it
 */
int spool_dump_buf_to_file(const void *buf, unsigned int buflen, const char *filepath) {
    FILE *fp;
    size_t bytes_written = 0;

    if ((fp = fopen(filepath, "wbx")) == NULL) {
        AGENT_LOG("Failed to open %s: %s", filepath, strerror(errno));
        return -1;
    }
        
    while (buflen > 0) {
        bytes_written = fwrite(buf, 1, buflen, fp);
        if (bytes_written < buflen && errno != EINTR) {
            fclose(fp);
            remove(filepath);
            return -1;
        }
        buflen -= bytes_written;
        buf += bytes_written;
    }
    fclose(fp);
    return 1;
}

/**
 * A wrapper that lets us override fopen for easier unit testing
 */
FILE *(*fopen_wrapper)(const char *path, const char *mode) = fopen;

/**
 * Store an incoming message in the spool
 *
 * @param msgid unique identifier for the message
 * @param buf pointer to the message contents
 * @param buflen length of message contents
 * @return -1 on error, 0 otherwise
 */
int spool_store(const char *msgid, const void *buf, unsigned int buflen) {
    char *dir, *msgpath;

    dir = spool_outgoing_dir();
    if (dir == NULL)
        return -1;

    if (spool_ensure_dir(dir) < 0) {
        free(dir);
        return -1;
    }

    msgpath = util_join_paths(dir, msgid);

    free(dir);

    if (spool_dump_buf_to_file(buf, buflen, msgpath) < 0) {
        free(msgpath);
        return -1;
    }

    free(msgpath);

    return 0;
}

int spool_watch(spool_callback cb) {
    char *incoming_dir;

    if (inotify_fd == -1) {
        inotify_fd = inotify_init();
        if (inotify_fd < 0)
            return -1;

        if (fcntl(inotify_fd, F_SETFL, O_NONBLOCK) < 0) {
            close(inotify_fd);
            inotify_fd = -1;
            return -1;
        }
    }

    incoming_dir = spool_incoming_dir();
    if (spool_ensure_dir(incoming_dir) < 0) {
        free(incoming_dir);
        return -1;
    }

    if (inotify_add_watch(inotify_fd, incoming_dir, IN_CLOSE_WRITE) < 0) {
        free(incoming_dir);
        perror("inotify_add_watch failed");
        return -1;
    }

    free(incoming_dir);

    if (slist_append(callbacks, cb) == NULL)
    {
        return -1;
    }

    return 0;
}

#define INOTIFY_BUFF_SIZE ((sizeof(struct inotify_event)+FILENAME_MAX))

void spool_process_pending_events(void) {
    ssize_t len, i;
    char buf[INOTIFY_BUFF_SIZE];

    while (1) {
        len = read(inotify_fd, buf, INOTIFY_BUFF_SIZE);
        if (len < 0) {
            /* (errno == EAGAIN && errno == EWOULDBLOCK) would mean that everything
             * fine, there just hasn't been any action, but we just return no matter
             * what */
            return;
        }

        i = 0;
        while (len >= sizeof(struct inotify_event)) {
            struct inotify_event *ev = (struct inotify_event *) &buf[i];

            len -= sizeof(struct inotify_event);

            if (ev->len) {
                char *path, *incoming_dir;
                SListElem *elem;

                if (ev->len > len) {
                   /* No room left in buffer, so exit out of this packet */
                   break;
                }

                incoming_dir = spool_incoming_dir();
                path = util_join_paths(incoming_dir, ev->name);
                free(incoming_dir);

                for(elem = slist_first(callbacks);
                        elem != NULL;
                        elem = slist_next(callbacks, elem)) {
                    spool_callback cb = (spool_callback)slist_data(elem);
                    cb(path);
                }

                free(path);

                len -= ev->len;
            }

            i += sizeof(struct inotify_event) + ev->len;
        }
    }
}
