/*
 *  Copyright 2010 Rackspace US, Inc.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"

/*
 * Initialise a strbuf "object"
 *
 * @returns: a pointer to a strbuf on success, NULL otherwise
 */
strbuf *strbuf_init() {
    strbuf *str;

    if ((str = malloc(sizeof(strbuf))) == NULL)
        return NULL;
    
    if ((str->str = strdup("")) == NULL) {
        free(str);
        return NULL;
    }

    str->len = 0;
    return str;
}

/*
 * Free a strbuf object, optionally returning the string it contained
 *
 * @param str: strbuf to be freed
 * @param free_it: if 0, the string contained in the strbuf is simply
 *                 freed, otherwise it is returned.
 */
char *strbuf_free(strbuf *str, char free_it) {
    char *s = str->str;

    free(str);

    if (free_it) {
        free(s);
        return NULL;
    } else {
        return s;
    }
}

strbuf *strbuf_append_printf(strbuf *str, const char *fmt, ...) {
    char *s = NULL;
    void *vptr;
    size_t strlen;
    va_list ap;

    va_start(ap, fmt);
    strlen = vsnprintf(s, 0, fmt, ap);
    va_end(ap);

    vptr = realloc(str->str, str->len + strlen + 1);
    if (vptr == NULL)
        return NULL;

    str->str = vptr;

    va_start(ap, fmt);
    vsnprintf(str->str+str->len, strlen+1, fmt, ap);
    va_end(ap);

    str->len+=strlen;

    return str;
}

/**
 * Join two paths
 *
 * @param part1 First part that should be joined
 * @param part2 Second part that should be joined
 * @return Newly allocated string: part1 + "/" + part2
 */
char *util_join_paths(const char *part1, const char *part2) {
    strbuf *str;
    strbuf *nstr;

    str = strbuf_init();
    if (str == NULL)
        return NULL;
    nstr = strbuf_append_printf(str, "%s/%s", part1, part2);
    if (nstr == NULL) {
        strbuf_free(str, 1);
        return NULL;
    }
    str = nstr;
    return strbuf_free(str, 0);
}
