/*
 *  Copyright 2010 Rackspace US, Inc.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
#include <sys/select.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <xs.h>
#include "log.h"
#include "slist.h"
#include "xen.h"

/*
 * Local types
 */
typedef int (*xen_cb)(const char *, const char *, const size_t, const char *);

typedef struct {
    char *token;
    xen_cb callback;
} token_callback_tuple;

/*
 * Global variables
 */
SList *handlers = NULL;
struct xs_handle *xh;

/*
 * Static function prototypes
 */
static void _xen_destroy_callback(void *data);

/*
 * Static functions
 */
static void _xen_destroy_callback(void *data) {
    token_callback_tuple *cb = (token_callback_tuple *)data;

    free(cb->token);
    free(cb);
}

void xen_fire_callback(const char *path, const char *token, void *buf,
                       unsigned int buflen);
int xen_convert_network_event(char *prefix, char **buf, unsigned int *buflen);

/*
 * Initialise the connection to Xen Store
 *
 * @returns 0 on success, -1 otherwise
 */
int xen_init(void) {
    if (!xh) {
        handlers = slist_init(_xen_destroy_callback);
        if (handlers == NULL)
            return -1;

        if (!(xh = xs_domain_open())) {
            AGENT_LOG("Failed to connect to Xen Store: %s", strerror(errno));
            slist_deinit(handlers);
        }
    }
    return xh ? 0 : -1;
}

/*
 * Register a watch with Xen Store.
 *
 * @param path: Path in Xen Store to monitor
 * @param callback: A xen_cb representing the callback to invoke when
 *                  the watched path is changed
 * @param data: A string to pass to the callback
 * @returns 0 on success, -1 otherwise.
 */
int xen_register_watch(const char *path,
                       xen_cb callback,
                       const char *data) {
    token_callback_tuple *cb;

    assert(xh != NULL); /* make sure xen_init() was called */

    if (!(cb = malloc(sizeof(token_callback_tuple)))) {
        AGENT_LOG("malloc: %s", strerror(errno));
        return -1;
    }

    AGENT_LOG("Starting to watch %s", path);
    if (!(xs_watch(xh, path, data))) {
        AGENT_LOG0("Failed to register watch with Xen Store");
        free(cb);
        return -1;
    }

    if (!(cb->token = strdup(data))) {
        AGENT_LOG("strdup failed: %s:", strerror(errno));
        free(cb);
        return -1;
    }

    cb->callback = callback;

    if (slist_append(handlers, cb) == NULL) {
        AGENT_LOG0("Failed to append callback to handler list.");
        free(cb->token);
        free(cb);
        return -1;
    }

    return 0;
}

#define NETWORKRESET_EVENT "{\"name\":\"resetnetwork\",\"value\":"
#define XENSTORE_NETWORKING_PATH "vm-data/networking"

/*
 * Pull the networking data out of a the xenstore, and return a new
 * buffer that contains 'prefix' appended with the networking data
 * Prefix is assumed to be of format:
 *    "{\"opt\":\"opt_value\",\"value\":"
 * The final '}' will be appended by this call.
 */
int xen_convert_network_event(char *prefix, char **buf, unsigned int *buflen)
{
    char dir_entry_path[255];
    char **dir_list;
    unsigned int dir_list_num;
    void *dir_entry;
    unsigned int dir_entry_len;
    char **d;
    unsigned int i;
    char *new_buf;
    unsigned int new_buf_len;
    void *vptr;

    dir_list = xen_read_dir(XENSTORE_NETWORKING_PATH, &dir_list_num);
    if (dir_list == NULL) {
        return -1;
    }

    new_buf_len = strlen(prefix);

    /* make sure there's room for starting '[', and ending ']' '}' and '\0' */
    new_buf = malloc(new_buf_len + 4);
    if (new_buf == NULL) {
        return -1;
    }

    strcpy(new_buf, prefix);
    new_buf[new_buf_len++] = '[';

    for(i=0,d=dir_list;i<dir_list_num;i++,d++)
    {
        snprintf(dir_entry_path, sizeof(dir_entry_path), "%s/%s",
                XENSTORE_NETWORKING_PATH, *d);

        dir_entry = xen_read_path(dir_entry_path, &dir_entry_len);
        if (dir_entry == NULL)
        {
            free(new_buf);
            free(dir_list);
            return -1;
        }

        /* make sure there's room for ending ',' ']' '}' and '\0' */
        vptr = realloc(new_buf, new_buf_len + dir_entry_len + 4);
        if (vptr == NULL) {
            free(dir_entry);
            free(new_buf);
            free(dir_list);
            return -1;
        }

        new_buf = vptr;

        if (i != 0)
            new_buf[new_buf_len++] = ',';

        memcpy(new_buf + new_buf_len, dir_entry, dir_entry_len);
        new_buf_len += dir_entry_len;

        free(dir_entry);
    }

    new_buf[new_buf_len++] = ']';
    new_buf[new_buf_len++] = '}';
    new_buf[new_buf_len] = '\0';

    free(dir_list);

    *buf = new_buf;
    *buflen = new_buf_len;

    return 0;
}

int xen_process_pending_events(void) {
    char **vec;
    unsigned int num;
    char *buf;
    unsigned int buflen;
    int fd;
    fd_set rfds;
    struct timeval timeout;
    int fds;

    if ((fd = xs_fileno(xh)) < 0) {
        AGENT_LOG("Failed to get file descriptor for connection to "
                  "Xen Store: %s", strerror(errno));
        return -1;
    }

    timeout.tv_sec = 0;
    timeout.tv_usec = 500000;

    FD_ZERO(&rfds);
    FD_SET(fd, &rfds);

   if ((fds = select(fd+1, &rfds, NULL, NULL, &timeout)) < 1)
       return 0;

    /*
     * Documentation on how this can fail is very scarce, so we'll
     * just keep our fingers crossed, hoping it'll never fail. :(
     */
    vec = xs_read_watch(xh, &num);
    if (num != 2) {
        free(vec);
        return -1;
    }

    AGENT_LOG("vec[XS_WATCH_PATH]: %s", vec[XS_WATCH_PATH]);
    if ((buf = xen_read_path(vec[XS_WATCH_PATH], &buflen)) == NULL) {
        free(vec);
        return -1;
    }

    if (!strncmp(buf, NETWORKRESET_EVENT, strlen(NETWORKRESET_EVENT))) {
        /*
         * A 'networkreset' event comes in with an empty value, so we
         * need to pull the networking data out of xenstore separately
         * and shove it in as the value
         */

        free(buf);
        if (xen_convert_network_event(NETWORKRESET_EVENT, &buf, &buflen) < 0) {
            free(vec);
            return -1;
        }
    }

    xen_fire_callback(vec[XS_WATCH_PATH], vec[XS_WATCH_TOKEN], buf, buflen);

    free(vec);
    free(buf);

    return 0;
}

int xen_write_path(const char *path, void *buf, unsigned int buflen) {
    xs_transaction_t t;
    int retval;
    int max_tries = 5;

    do {
        if (!(t = xs_transaction_start(xh))) {
            AGENT_LOG0("Failed to start transaction with Xen Store.");
            return -1;
        }
        retval = xs_write(xh, t, path, buf, buflen) ? -1 : 0;
        if (!xs_transaction_end(xh, t, 0)) {
            if (errno == EAGAIN && max_tries-- > 0)
                continue;
            else
                return -1;
        } else
            break;
    } while (1);
    return retval;
}

void xen_fire_callback(const char *path, const char *token, void *buf,
                            unsigned int buflen) {
    SListElem *elem;

    for (elem = slist_first(handlers);
             elem != NULL;
             elem = slist_next(handlers, elem)) {
        token_callback_tuple *t = (token_callback_tuple *)slist_data(elem);
        if (strcmp(token, t->token) == 0) {
            t->callback(path, buf, buflen, t->token);
            break;
        }
    }
}

void *xen_read_path(const char *path, unsigned int *buflen) {
    void *buf;
    xs_transaction_t t;
    int max_tries = 5;

    do {
        t = xs_transaction_start(xh);
        if (!(buf = xs_read(xh, t, path, buflen))) {
            /* Documentation on failure modes is lacking, so we
             * just end the transaction (ignoring failure) and
             * bail out. */
            xs_transaction_end(xh, t, 0);
            return NULL;
        }
        /* I have no clue what it means if ending the transaction
         * fails, even though xs_read succeeded. The docs say
         * we should try again, so let's just do that. */
        if (!xs_transaction_end(xh, t, 0)) {
            if (errno == EAGAIN && max_tries-- > 0)
                continue;
            else
                return NULL;
        } else
            break;
    } while (1);

    return buf;
}

char **xen_read_dir(const char *path, unsigned int *num_entries)
{
    char **entries;
    xs_transaction_t t;
    int max_tries = 5;

    do {
        t = xs_transaction_start(xh);
        if (!(entries = xs_directory(xh, t, path, num_entries))) {
            /* Documentation on failure modes is lacking, so we
             * just end the transaction (ignoring failure) and
             * bail out. */
            xs_transaction_end(xh, t, 0);
            return NULL;
        }
        /* I have no clue what it means if ending the transaction
         * fails, even though xs_read succeeded. The docs say
         * we should try again, so let's just do that. */
        if (!xs_transaction_end(xh, t, 0)) {
            if (errno == EAGAIN && max_tries-- > 0)
                continue;
            else
                return NULL;
        } else
            break;
    } while (1);

    return entries;
}
