/*
 *  Copyright 2010 Rackspace US, Inc.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <check.h>
#include <xs.h>

extern char xs_daemon_open_should_fail;

START_TEST(test_xs_daemon_open_failure)
{
	xs_daemon_open_should_fail = 1;
    fail_unless((xs_daemon_open()) == NULL, "xs_daemon_open succeeded, but we asked it to fail!");
}
END_TEST

START_TEST(test_xs_daemon_open)
{
    fail_unless((xs_daemon_open()) != NULL, "xs_daemon_open failed!");
}
END_TEST

START_TEST(test_xs_transaction_start_end)
{
	struct xs_handle *h;
	xs_transaction_t t1, t2;
	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	fail_unless((t1 = xs_transaction_start(h)) != 0, "xs_transaction_start did not return a transaction handle");
	fail_unless((t2 = xs_transaction_start(h)) != 0, "xs_transaction_start did not return a transaction handle the second time");
	fail_unless(t1 != t2, "xs_transaction_start returned identical xs_transaction_t's");
	fail_unless(xs_transaction_end(h, t1, 0), "xs_transaction_end failed");
	fail_unless(xs_transaction_end(h, t2, 0), "xs_transaction_end failed");
}
END_TEST

START_TEST(test_xs_transaction_start_invalid_xs_handle)
{
	struct xs_handle *h = NULL;
	xs_transaction_t t;
	fail_unless((t = xs_transaction_start(h)) == 0, "xs_transaction_start succeeded even though we gave an invalid struct xs_handle");
}
END_TEST

START_TEST(test_xs_transaction_end_invalid_xs_handle)
{
	struct xs_handle *h = NULL;
	xs_transaction_t t = 1;
	fail_if(xs_transaction_end(h, t, 0), "xs_transaction_end succeeded even though we gave an invalid struct xs_handle");
}
END_TEST

START_TEST(test_xs_transaction_end_invalid_xs_transaction)
{
	struct xs_handle *h;
	xs_transaction_t t = 0;
	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	fail_if(xs_transaction_end(h, t, 0), "xs_transaction_end succeeded even though we gave an invalid xs_transaction_t");
}
END_TEST

START_TEST(test_xs_read_invalid_xs_handle) {
	struct xs_handle *h = NULL;
	xs_transaction_t t = 1;
	unsigned int len;

	fail_unless(xs_read(h, t, "", &len) == NULL, "xs_read succeeded even though we gave an invalid xs_handle");
}
END_TEST

START_TEST(test_xs_read_invalid_xs_transaction) {
	struct xs_handle *h;
	xs_transaction_t t = 1;
	unsigned int len;

	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	fail_unless(xs_read(h, t, "", &len) == NULL, "xs_read succeeded even though we gave an invalid xs_transaction_t");
}
END_TEST

START_TEST(test_xs_read_unknown_path) {
	struct xs_handle *h;
	xs_transaction_t t;
	unsigned int len;

	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	fail_unless((t = xs_transaction_start(h)) != 0, "xs_transaction_start did not return a transaction handle");
	fail_unless((xs_read(h, t, "/unknown/path", &len) == 0), "xs_read of unknown path gave successful return value");
}
END_TEST

START_TEST(test_xs_read) {
	struct xs_handle *h;
	xs_transaction_t t;
	unsigned int len;
	void *buf;

	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	fail_unless((t = xs_transaction_start(h)) != 0, "xs_transaction_start did not return a transaction handle");
	fail_unless((xs_write(h, t, "/somedir/somefile", "contents", 8) == 1), "xs_write failed");
	fail_unless((buf = xs_read(h, t, "/somedir/somefile", &len)) != 0, "xs_read failed");
	fail_unless(len == 8, "Invalid length data returned");
	fail_unless(memcmp(buf, "contents", 8) == 0, "Invalid data returned");
	fail_unless((buf = xs_read(h, t, "/somedir/somefile", &len)) != 0, "xs_read failed");
	fail_unless((xs_read(h, t, "/unknown/path", &len) == 0), "xs_read of unknown path gave successful return value");
}
END_TEST

START_TEST(test_xs_write_invalid_xs_handle) {
	struct xs_handle *h = NULL;
	xs_transaction_t t = 1;

	fail_unless(xs_write(h, t, "", "", 0) == 0, "xs_write succeeded even though we gave an invalid xs_handle");
}
END_TEST

START_TEST(test_xs_write_invalid_xs_transaction) {
	struct xs_handle *h;
	xs_transaction_t t = 1;

	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	fail_unless(xs_write(h, t, "", "", 0) == 0, "xs_write succeeded even though we gave an invalid xs_transaction_t");
}
END_TEST

START_TEST(test_xs_write) {
	struct xs_handle *h;
	xs_transaction_t t;

	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	fail_unless((t = xs_transaction_start(h)) != 0, "xs_transaction_start did not return a transaction handle");
	fail_unless((xs_write(h, t, "/somedir/somefile", "contents", 8) == 1), "xs_write failed");
}
END_TEST

START_TEST(test_xs_write_twice_overwrites) {
	struct xs_handle *h;
	xs_transaction_t t;
	void *buf;
	unsigned int len;

	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	fail_unless((t = xs_transaction_start(h)) != 0, "xs_transaction_start did not return a transaction handle");
	fail_unless((xs_write(h, t, "/somedir/somefile", "contents", 8) == 1), "xs_write failed");
	fail_unless((xs_write(h, t, "/somedir/somefile", "otherstuff", 10) == 1), "xs_write failed");
	fail_unless((buf = xs_read(h, t, "/somedir/somefile", &len)) != 0, "xs_read failed");
	fail_unless(len == 10, "Invalid length data returned");
	fail_unless(memcmp(buf, "otherstuff", 10) == 0, "Invalid data returned");
}
END_TEST

START_TEST(test_xs_watch_invalid_xs_handle) {
	struct xs_handle *h = NULL;
	fail_unless(xs_watch(h, "", "") == 0, "xs_watch succeeded even though we gave an invalid xs_handle");
}
END_TEST

START_TEST(test_xs_watch) {
	struct xs_handle *h;

	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	fail_unless(xs_watch(h, "", "") != 0, "xs_watch failed");
}
END_TEST

START_TEST(test_xs_read_watch_invalid_xs_handle) {
	struct xs_handle *h = NULL;
	unsigned int len;

	fail_unless(xs_read_watch(h, &len) == NULL, "xs_read_watch succeeded even though we gave an invalid xs_handle");
}
END_TEST

extern int nonblocking;

START_TEST(test_xs_read_watch_nothing_pending) {
	struct xs_handle *h;
	char **vec;
	unsigned int len;

	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	nonblocking = 1;
	fail_unless((vec = xs_read_watch(h, &len)) == NULL &&
	             errno == EWOULDBLOCK,
				"xs_read_watch didn't say that it would have blocked.");
}
END_TEST

START_TEST(test_xs_read_watch) {
	struct xs_handle *h;
	xs_transaction_t t;
	unsigned int len;
	char **vec;

	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	fail_unless(xs_watch(h, "/watched", "mytoken") != 0, "xs_watch failed");
	fail_unless((t = xs_transaction_start(h)) != 0, "xs_transaction_start did not return a transaction handle");
	fail_unless((xs_write(h, t, "/watched/file", "contents", 8) == 1), "xs_write failed");
	fail_unless(xs_transaction_end(h, t, 0), "xs_transaction_end failed");
	fail_unless((vec = xs_read_watch(h, &len)) != 0, "xs_read_watch failed");
    fail_unless(strcmp(vec[XS_WATCH_PATH], "/watched/file") == 0, "Wrong path returned from xs_read_watch");
    fail_unless(strcmp(vec[XS_WATCH_TOKEN], "mytoken") == 0, "Wrong token returned from xs_read_watch");
}
END_TEST

START_TEST(test_xs_fileno_invalid_xs_handle) {
	struct xs_handle *h = NULL;

	fail_unless(xs_fileno(h) == -1, "xs_fileno gave a valid fd even though we gave an invalid xs_handle");
}
END_TEST

START_TEST(test_xs_fileno) {
	struct xs_handle *h;

	fail_unless((h = xs_daemon_open()) != NULL, "xs_daemon_open failed");
	fail_unless(xs_fileno(h) >= 0, "xs_fileno failed");
}
END_TEST

Suite *mock_xenstore_suite(void)
{
    Suite *s = suite_create("Mock Xenstore");

    TCase *tc_xs_daemon_open = tcase_create("xs_daemon_open");
    tcase_add_test(tc_xs_daemon_open, test_xs_daemon_open);
    tcase_add_test(tc_xs_daemon_open, test_xs_daemon_open_failure);
    suite_add_tcase(s, tc_xs_daemon_open);

    TCase *tc_xs_transaction = tcase_create("xs_transaction");
    tcase_add_test(tc_xs_transaction, test_xs_transaction_start_end);
    tcase_add_test(tc_xs_transaction, test_xs_transaction_start_invalid_xs_handle);
    tcase_add_test(tc_xs_transaction, test_xs_transaction_end_invalid_xs_handle);
    tcase_add_test(tc_xs_transaction, test_xs_transaction_end_invalid_xs_transaction);
    suite_add_tcase(s, tc_xs_transaction);

    TCase *tc_xs_write = tcase_create("xs_write");
    tcase_add_test(tc_xs_write, test_xs_write_invalid_xs_handle);
    tcase_add_test(tc_xs_write, test_xs_write_invalid_xs_transaction);
    tcase_add_test(tc_xs_write, test_xs_write);
    tcase_add_test(tc_xs_write, test_xs_write_twice_overwrites);
    suite_add_tcase(s, tc_xs_write);

    TCase *tc_xs_read = tcase_create("xs_read");
    tcase_add_test(tc_xs_read, test_xs_read_invalid_xs_handle);
    tcase_add_test(tc_xs_read, test_xs_read_invalid_xs_transaction);
    tcase_add_test(tc_xs_read, test_xs_read_unknown_path);
    tcase_add_test(tc_xs_read, test_xs_read);
    suite_add_tcase(s, tc_xs_read);

    TCase *tc_xs_watch = tcase_create("xs_watch");
    tcase_add_test(tc_xs_watch, test_xs_watch_invalid_xs_handle);
    tcase_add_test(tc_xs_watch, test_xs_watch);
    suite_add_tcase(s, tc_xs_watch);

    TCase *tc_xs_read_watch = tcase_create("xs_read_watch");
    tcase_add_test(tc_xs_read_watch, test_xs_read_watch_invalid_xs_handle);
    tcase_add_test(tc_xs_read_watch, test_xs_read_watch);
    tcase_add_test(tc_xs_read_watch, test_xs_read_watch_nothing_pending);
    suite_add_tcase(s, tc_xs_read_watch);

    TCase *tc_xs_fileno = tcase_create("xs_fileno");
    tcase_add_test(tc_xs_fileno, test_xs_fileno_invalid_xs_handle);
    tcase_add_test(tc_xs_fileno, test_xs_fileno);
    suite_add_tcase(s, tc_xs_fileno);

    return s;
}

int main(void)
{
    int number_failed;

    Suite *s = mock_xenstore_suite();
    SRunner *sr = srunner_create(s);
    srunner_run_all(sr, CK_NORMAL);
    number_failed = srunner_ntests_failed(sr);
    srunner_free(sr);
    return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
