/*
 *  Copyright 2010 Rackspace US, Inc.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <check.h>
#include "../src/spool.h"
#include "../src/util.h"
#include "../config.h"

#ifndef SPOOLDIR
#define SPOOLDIR "/var/spool/agent"
#endif

extern char *spool_basedir(void);
extern char *spool_incoming_dir(void);
extern char *spool_outgoing_dir(void);

START_TEST(test_spool_basedir)
{
    unsetenv("AGENT_MSG_SPOOL");
    fail_unless(strcmp(spool_basedir(), SPOOLDIR) == 0, "spool_basedir did not return SPOOLDIR");
    setenv("AGENT_MSG_SPOOL", "foobar", 1);
    fail_unless(strcmp(spool_basedir(), "foobar") == 0, "spool_basedir did not return \"foobar\"");
}
END_TEST

START_TEST(test_spool_incoming_dir)
{
    setenv("AGENT_MSG_SPOOL", "foobar", 1);
    fail_unless(strcmp(spool_incoming_dir(), "foobar/incoming") == 0, "spool_basedir didn not return \"foobar/incoming\"");
}
END_TEST

START_TEST(test_spool_outgoing_dir)
{
    setenv("AGENT_MSG_SPOOL", "foobar", 1);
    fail_unless(strcmp(spool_outgoing_dir(), "foobar/outgoing") == 0, "spool_basedir didn not return \"foobar/outgoing\"");
}
END_TEST

extern int spool_ensure_dir(const char *dir);

char testdir[255];

void test_spool_store_setup(void) {
    strcpy(testdir, "myspoolXXXXXX");
    fail_unless(mkdtemp((char *)testdir) != NULL);
}

void test_spool_store_teardown(void) {
	char cmd[1024];
	int ret;

    snprintf(cmd, 1024, "rm -rf %s", testdir);
    ret = system(cmd);
	if (ret < 0 || WEXITSTATUS(ret) != 0)
		fprintf(stderr, "Failed to remove temporary directory %s", testdir);

}

START_TEST(test_spool_ensure_dir)
{
    struct stat dummy;

    remove(testdir);
    spool_ensure_dir(testdir);
    fail_unless(stat(testdir, &dummy) == 0);
}
END_TEST

START_TEST(test_spool_dumpbuf_to_file)
{
    char msgfile[255], buf[100];
    FILE *fp;

    setenv("AGENT_MSG_SPOOL", testdir, 1);
    spool_store("some_msg_id", "foobar", 6);

    snprintf(msgfile, 255, "%s/outgoing/some_msg_id", testdir);

    fail_unless((fp = fopen(msgfile, "r")) != NULL, "Failed to open the spool file. Perhaps it never got created?");
    fail_unless(fread((void *) buf, 1, 100, fp) == 6, "Incorrect amount of data in the spool file");
    fail_unless(memcmp(buf, "foobar", 6) == 0, "Spool file did not contain the correct data");
}
END_TEST

char spool_cb_called = 0;

void spool_cb(const char *path) {
    spool_cb_called = 1;

}

START_TEST(test_spool_watch) {
    FILE *fp;

    setenv("AGENT_MSG_SPOOL", testdir, 1);
    fail_unless(spool_init() == 0, "Failed to init spool");
    fail_unless(spool_watch(spool_cb) == 0, "Failed to add watch.");
    spool_process_pending_events();
    fail_unless(spool_cb_called == 0, "Spool callback got called even though nothing had happened yet.");
    fp = fopen(util_join_paths(spool_incoming_dir(), "testfile"), "a+");
    fwrite("testing", 1, 8, fp);
    fclose(fp);
    spool_process_pending_events();
    fail_unless(spool_cb_called == 1, "Spool callback didn't get called at all.");
}
END_TEST

Suite *spool_suite(void)
{
    Suite *s = suite_create("Spool");

    TCase *tc_spool_store = tcase_create("spool_store");
    tcase_add_checked_fixture(tc_spool_store, test_spool_store_setup, test_spool_store_teardown);
    tcase_add_test(tc_spool_store, test_spool_basedir);
    tcase_add_test(tc_spool_store, test_spool_incoming_dir);
    tcase_add_test(tc_spool_store, test_spool_outgoing_dir);
    tcase_add_test(tc_spool_store, test_spool_ensure_dir);
    tcase_add_test(tc_spool_store, test_spool_dumpbuf_to_file);
    tcase_add_test(tc_spool_store, test_spool_watch);
    suite_add_tcase(s, tc_spool_store);

    return s;
}

int main(void)
{
    int number_failed;

    Suite *s = spool_suite();
    SRunner *sr = srunner_create(s);
    srunner_run_all(sr, CK_NORMAL);
    number_failed = srunner_ntests_failed(sr);
    srunner_free(sr);
    return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
