/*
 *  Copyright 2010 Rackspace US, Inc.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <check.h>
#include <xs.h>
#include "../src/xen.h"

#define test_read_path "/somepath/something"
#define test_watch_path "/somepath"
#define test_token "somedata"
#define test_data "some\0random\0date"

extern char xs_daemon_open_should_fail;

START_TEST(test_xen_init_success)
{
    fail_if(xen_init(), "xen_init failed!");
}
END_TEST

START_TEST(test_xen_init_failure)
{
	xs_daemon_open_should_fail = 1;
    fail_unless(xen_init(), "xen_init succeeded, but should've failed!");
}
END_TEST

START_TEST(test_xen_write_and_read_path)
{
	unsigned int buflen;
	void *buf;

    xen_init();
	xen_write_path(test_read_path, test_data, sizeof(test_data));
	buf = xen_read_path(test_read_path, &buflen);
    fail_unless(buflen == sizeof(test_data),
                "xen_read_path altered the length of the data buffer");
    fail_unless(memcmp(buf, test_data, buflen) == 0,
                "xen_read_path mangled data buffer");
}
END_TEST

/*
 * Bookkeeping
 */
static int testcb_called = 0;

static int testcb(const char *path, const char *buf, size_t buflen, const char *data) {
    fail_unless(strcmp(path,
                       "/somepath/something") == 0,
                "Data was at /somepath/something, but that's not what "
                "our callback was told.");
    fail_unless(buflen == sizeof(test_data),
                "Callback was called with an incorrectly sized size "
                "buffer");
    fail_unless(memcmp(buf, test_data, buflen) == 0,
                "Buffer passed to callback did not match the original "
                "data.");
    fail_unless(strcmp(data, test_token) == 0, "Token did not match");
    testcb_called = 1;
	return 0;
}

START_TEST(test_xen_fire_callback)
{
    xen_init();
	xen_register_watch(test_watch_path, testcb, test_token);
	xen_write_path(test_read_path, test_data, sizeof(test_data));
	xen_process_pending_events();
    fail_unless(testcb_called, "Callback didn't fire at all.");
}
END_TEST

Suite *xen_suite(void)
{
    Suite *s = suite_create("Xen");

    TCase *tc_xen_init = tcase_create("xen_init");
    tcase_add_test(tc_xen_init, test_xen_init_success);
    tcase_add_test(tc_xen_init, test_xen_init_failure);
    suite_add_tcase(s, tc_xen_init);

    TCase *tc_xen_write_and_read_path = tcase_create("xen_write_and_read_path");
    tcase_add_test(tc_xen_write_and_read_path, test_xen_write_and_read_path);
    suite_add_tcase(s, tc_xen_write_and_read_path);

    TCase *tc_xen_fire_callback = tcase_create("xen_fire_callback");
    tcase_add_test(tc_xen_fire_callback, test_xen_fire_callback);
    suite_add_tcase(s, tc_xen_fire_callback);

    return s;
}

int main(void)
{
    int number_failed;

    Suite *s = xen_suite();
    SRunner *sr = srunner_create(s);
    srunner_run_all(sr, CK_NORMAL);
    number_failed = srunner_ntests_failed(sr);
    srunner_free(sr);
    return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
