/*
 *  Copyright 2010 Rackspace US, Inc.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
#include <malloc.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "../src/slist.h"
#include <xs.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>

struct xs_handle {
    SList *xs_object_list;
    SList *watches;
    SList *pending_notifications;
    char *verification;
    char transactions;
    int clientfd;
    int libraryfd;
};

struct xs_object {
    char *path;
    void *buf;
    unsigned int buflen;
};

struct xs_path_watch {
    const char *path;
    const char *token;
};

struct notification {
    const char *path;
    const char *token;
};

static int validate_xs_handle(struct xs_handle *h) {
    if (h && !strcmp(h->verification, "foo"))
        return 1;
    return 0;
}

static int validate_xs_transaction(struct xs_handle *h, xs_transaction_t t) {
    if (h && !strcmp(h->verification, "foo") && (h->transactions & (1 << t)))
        return 1;
    return 0;
}

char xs_daemon_open_should_fail = 0;

struct xs_handle *xs_domain_open(void) {
	return xs_daemon_open();
}

struct xs_handle *xs_daemon_open(void) {
    int fds[2];

    if (xs_daemon_open_should_fail)
        return NULL;

    struct xs_handle *retval = malloc(sizeof(struct xs_handle));

    retval->watches = slist_init(NULL);
    retval->xs_object_list = slist_init(NULL);
    retval->pending_notifications = slist_init(NULL);

    retval->verification = strdup("foo");
    retval->transactions = 0;

    socketpair(AF_UNIX, SOCK_STREAM, 0, fds);
    retval->clientfd = fds[0]; /* The library consumer end */
    retval->libraryfd = fds[1]; /* Our end */
    return retval;
}

xs_transaction_t xs_transaction_start(struct xs_handle *h) {
    xs_transaction_t t = 1;

    if (!validate_xs_handle(h)) 
        return 0;

    /* Find a free transaction ID */
    while ((h->transactions >> t) & 1)
        t++;

    /* Mark it used */
    h->transactions |= 1 << t;

    return t;
}

bool xs_transaction_end(struct xs_handle *h, xs_transaction_t t,
                        bool abort) {
    if (!validate_xs_transaction(h, t))
        return 0;

    /* Mark it free again */
    h->transactions &= ~(1 << t);

    return 1;
}

bool xs_write(struct xs_handle *h, xs_transaction_t t, const char *path, const void *buf, unsigned int len) {
    SListElem *ptr;
    struct xs_object *obj;

    if (!validate_xs_transaction(h, t))
        return 0;

    for (ptr = slist_first(h->watches);
             ptr != NULL;
             ptr = slist_next(h->watches, ptr)) {
        struct xs_path_watch *watch = (struct xs_path_watch *)slist_data(ptr);
        if (strstr(path, watch->path) == path) {
            /* Either they must be identical or the triggering path be in a subdir of the watched path */
            if (strlen(path) == strlen(watch->path) || path[strlen(watch->path)] == '/') {
                struct notification *n = (struct notification *) malloc(sizeof(struct notification));
                n->path = strdup(path);
                n->token = strdup(watch->token);
                slist_append(h->pending_notifications, n);
                /* We notify the client that there's some action */
                if (write(h->libraryfd, "X", 1) < 1)
                    abort();
            }
        }
    }

    for (ptr = slist_first(h->xs_object_list);
             ptr != NULL;
             ptr = slist_next(h->xs_object_list, ptr)) {
        obj = (struct xs_object *)slist_data(ptr);
        if (strcmp(obj->path, path) == 0) {
            free(obj->buf);
            obj->buf = malloc(len);
            memcpy(obj->buf, buf, len);
            obj->buflen = len;
            return 1;
        }
    }
    obj = (struct xs_object *) malloc(sizeof(struct xs_object));
    obj->path = strdup(path);
    obj->buf = malloc(len);
    memcpy(obj->buf, buf, len);
    obj->buflen = len;
    slist_append(h->xs_object_list, obj);

    return 1;
}

void *xs_read(struct xs_handle *h, xs_transaction_t t, const char *path, unsigned int *len) {
    SListElem *ptr;

    if (!validate_xs_transaction(h, t))
        return NULL;

    for (ptr = slist_first(h->xs_object_list);
             ptr != NULL;
             ptr = slist_next(h->xs_object_list, ptr)) {
        struct xs_object *obj = (struct xs_object *)slist_data(ptr);

        if (strcmp(obj->path, path) == 0) {
            void *buf;

            buf = malloc(obj->buflen);
            memcpy(buf, obj->buf, obj->buflen);
            *len = obj->buflen;
            return buf;
        }
    }
    return NULL;
}

bool xs_watch(struct xs_handle *h, const char *path, const char *token) {
    struct xs_path_watch *watch;

    if (!validate_xs_handle(h)) 
        return 0;

    watch = (struct xs_path_watch *) malloc(sizeof(struct xs_path_watch));
    watch->path = strdup(path);
    watch->token = strdup(token);
    slist_append(h->watches, watch);

    return 1;
}

int nonblocking = 0;

char **xs_read_watch(struct xs_handle *h, unsigned int *num) {
    char **retval;
    SListElem *elem;

    if (!validate_xs_handle(h))
        return NULL;

    if ((elem = slist_first(h->pending_notifications)) == NULL) {
        if (nonblocking) {
            errno = EWOULDBLOCK;
            return NULL;
        } else {
            fd_set rfds;

            FD_ZERO(&rfds);
            FD_SET(h->clientfd, &rfds);

            select(h->clientfd+1, &rfds, NULL, NULL, NULL);
        }
    }

    retval = (char **) malloc(sizeof(char *)*2);
    *retval = strdup(((struct notification *) slist_data(elem))->path);
    *(retval + 1) = strdup(((struct notification *) slist_data(elem))->token);
    *num = 2;

    return retval;
}

int xs_fileno(struct xs_handle *h) {
    if (!validate_xs_handle(h))
        return -1;

    return h->clientfd;
}
